"""Apply a vocab mapping converting a list of tokens in a column
into a list of integers"""

from abc import abstractmethod
from math import ceil
from sys import getsizeof
from typing import List, Union

import numpy as np
import polars as pl
import pyarrow as pa

from bardi.data.utils.pyarrow_utils import chunk_pyarrow_table
from bardi.nlp_engineering.utils.validations import (
    validate_list_str_cols,
    validate_pyarrow_table,
)
from bardi.pipeline import DataWriteConfig, Step


class PostProcessor(Step):
    """The post processor maps a vocab to a list of tokens

    Note
    ----

    Avoid the direct instantiation of the PostProcessor class
    and instead instantiate one of the child classes depending
    on hardware configuration

    Attributes
    ----------

    fields : Union[str, List[str]]
        the name of the column containing a list of tokens
        that will be mapped to integers using a vocab
    field_rename : str
        optional ability to rename the supplied field with
        the field_rename value
    id_to_token : dict
        optional vocabulary in the form of {id: token} that
        will be used to map the tokens to integers. This
        is optional for the construction of the object, and
        can alternatively be provided in the run method.
        This flexibility handles the use of a pre-existing
        vocab versus creating a vocab during a pipeline run.
    concat_fields : bool
        indicate if you would like for fields to be concatenated
        into a single column or left as separate columns
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        field_rename: str = None,
        id_to_token: dict = None,
        concat_fields: bool = False,
    ):
        """Constructor method
        """
        self.fields = fields
        if isinstance(fields, str):
            self.fields = [fields]

        self.field_rename = "X"
        if field_rename:
            self.field_rename = field_rename

        self.mapping = None
        if id_to_token:
            # the actual application of the mapping requires
            # flipping the keys and values to have {token: id}
            self.mapping = {token: int(id) for id, token in id_to_token.items()}

        self.concat_fields = concat_fields
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config

    @abstractmethod
    def run(self):
        """Abstract method
        """
        pass


class CPUPostProcessor(PostProcessor):
    """The post processor maps a vocab to a list of tokens

    Note
    ----

    This implementation of the PostProcessor is specific for CPU computation.

    Attributes
    ----------

    fields : Union[str, List[str]]
        the name of the column containing a list of tokens
        that will be mapped to integers using a vocab
    field_rename : str
        optional ability to rename the supplied field with
        the field_rename value
    id_to_token : dict
        optional vocabulary in the form of {id: token} that
        will be used to map the tokens to integers. This
        is optional for the construction of the object, and
        can alternatively be provided in the run method.
        This flexibility handles the use of a pre-existing
        vocab versus creating a vocab during a pipeline run.
    concat_fields : bool
        indicate if you would like for fields to be concatenated
        into a single column or left as separate columns
    """

    def __init__(self, *args, **kwargs):
        """Constructor method
        """
        super().__init__(*args, **kwargs)

    def run(
        self,
        data: pa.Table,
        artifacts: dict = None,
        id_to_token: dict = None,
    ) -> pa.Table:
        """Run a post processor using CPU computation

        The post processor relies on receiving a vocab to map. The
        vocab can be supplied in multiple ways:
          - id_to_token at object creation
          - contained in the pipeline artifacts dictionary passed
            to the run method referenced by the key, 'id_to_token'
          - id_to_token in the run method

        Attributes
        ----------

        data : PyArrow Table
            The data to be processed. The data must contain the
            column specified by field at object creation
        artifacts : dict
            A dictionary of pipeline artifacts which contains
            a vocab referenced by the key, 'id_to_token'
        id_to_token : dict
            If a vocab wasn't passed at object creation or
            through the pipeline artifacts dict, then it
            must be passed here as a final option

        Returns
        -------

        Tuple(PyArrow Table, dict)
            A tuple with:
                - the first element holding a PyArrow Table of data
                processed with the post processor
                - the second element of the tuple intended for
                artifacts is None

        Raises
        ------

        AttributeError
            The vocab (id_to_token) wasn't supplied
            either at object creation or to the run method
        TypeError
            The run method was not supplied a PyArrow Table
        """

        # Perform validations
        validate_pyarrow_table(data=data)
        validate_list_str_cols(fields=self.fields, data=data)

        # Check if vocab was passed through the artifacts dict
        if "id_to_token" in artifacts.keys():
            id_to_token = artifacts["id_to_token"]

        # Check if a vocab was passed directly to the method -
        # the route used if a vocab was created during script execution
        if id_to_token:
            self.mapping = {token: id for id, token in id_to_token.items()}
        # Check if a vocab was referenced in the initial object creation -
        # the route used if there was a pre-existing vocab
        elif not self.mapping:
            # If a vocab wasn't referenced in either place, then there
            # isn't a vocab to use in this run method
            raise AttributeError(
                "id_to_token must be provided to either "
                "the PostProcessor object instantiation "
                "or to the run method"
            )

        # Grab the <unk> id from the supplied vocab
        try:
            self.unk_id = self.mapping["<unk>"]
        except KeyError:
            self.unk_id = None  # If there isn't an unk id

        # Map tokens to ids using the supplied vocab.
        # If a token encountered isn't in the vocab, it defaults to
        # the <unk> value. The field passed is renamed to 'X'

        # This step is memory intensive, so the data is split into
        # 128 MB chunks and operated upon to avoid OoM errors
        data_size = getsizeof(data) * 1e-6
        chunks = ceil(data_size / 128)
        column_order = data.column_names
        tables = chunk_pyarrow_table(
            data=data, row_count=data.num_rows, min_batches=chunks
        )

        self.mapping[None] = None

        for i, table in enumerate(tables):
            chunk_df = pl.from_arrow(table)
            # create a unique id for each row in the chunk
            # this is needed for exploding and grouping
            # the data for efficinet execution
            index_list = np.arange(0, chunk_df.height)
            index_col = pl.Series(index_list)
            chunk_df = chunk_df.with_columns(index_col.alias("unique_id"))

            for field in self.fields:
                meta_df = chunk_df.drop(field)
                tmp_df = (
                    chunk_df.select(["unique_id", field])
                    .explode(field)
                    .with_columns(
                        pl.col(field).map_dict(
                            self.mapping, default=self.unk_id, return_dtype=pl.Int64()
                        )
                    )
                    .group_by("unique_id", maintain_order=True)
                    .agg(pl.col(field))
                )

                chunk_df = meta_df.join(tmp_df, on="unique_id", how="left")
            if i == 0:
                df = chunk_df.select(*column_order)
            else:
                df = df.vstack(chunk_df.select(*column_order))

        # concat the fields if desired
        if self.concat_fields and len(self.fields) > 1:
            # Concatenate the values in the fields.
            # Fields at this point are assumed to be large lists

            current_field = self.fields[0]
            for i in range(len(self.fields) - 1):
                next_field = self.fields[i + 1]
                df = df.with_columns(
                    pl.col(current_field)
                    .list.concat(next_field)
                    .list.drop_nulls()
                    .alias("X")
                ).drop(next_field)
                current_field = "X"
            df = df.drop(self.fields[0])

            # concat field is called 'X', rename if
            # other name provided, defaults to 'X'
            df = df.rename({"X": self.field_rename})

        # if concat was not run and a single field was provided
        # the field can be renamed
        elif len(self.fields) == 1:
            df = df.rename({self.fields[0]: self.field_rename})

        data = df.to_arrow()

        return (data, None)

    def get_parameters(self):
        """Retrive the post-processor object configuration
        Does not return the mapping (vocab) as it can be large

        Returns
        -------

        dict
            a dictionary representation of the post-processor's attributes
        """
        params = vars(self).copy()
        params.pop("mapping")
        return params



"""Train a Word2Vec model and create a vocab and word embeddings"""

import os
from abc import abstractmethod
from multiprocessing import cpu_count
from os.path import isfile
from typing import List, TypedDict, Union, Tuple, Optional

import numpy as np
import polars as pl
import pyarrow as pa
from gensim.utils import RULE_DEFAULT, RULE_DISCARD
from gensim.models import Word2Vec

from bardi.data import data_handlers
from bardi.nlp_engineering.utils.validations import (
    validate_pyarrow_table,
    validate_list_str_cols,
)
from bardi.pipeline import DataWriteConfig, Step


class EmbeddingGeneratorArtifactsWriteConfig(TypedDict):
    """Indicates the keys and data types expected in an
    artifacts write config dict for the embedding generator
    if overwriting the default configuration."""

    vocab_format: str
    vocab_format_args: dict
    embedding_matrix_format: str
    embedding_matrix_format_args: dict


class EmbeddingGenerator(Step):
    """The embedding generator provides an interface to create
    word embeddings or vector representations of words (tokens).

    The embedding generator uses the Word2Vec model from the Gensim library.

    Note
    ----

    Avoid the direct instantiation of the PreTokenizer class
    and instead instantiate one of the child classes depending
    on hardware configuration.

    Attributes
    ----------

    fields : Union[str, List[str]]
        The name of the column(s) containing text to generate embeddings.
    load_saved_model : bool
        Whether to load a saved Word2Vec model or train a new one.
    checkpoint_path : str
        Path to the saved model checkpoint if `load_saved_model` is True.
    cores : int
        Number of CPU cores to use for training.
    min_word_count : int
        Ignore all words with a total frequency lower than this.
    window : int
        Maximum distance between the current and predicted word within a sentence.
    vector_size : int
        Dimensionality of the word vectors.
    sample : float
        The threshold for configuring which higher-frequency words are randomly downsampled.
    min_alpha : float
        Learning rate will linearly drop to `min_alpha` as training progresses.
    negative : int
        If > 0, specifies how many "noise words" should be drawn.
    epochs : int
        Number of iterations (epochs) over the corpus.
    seed : int
        Seed for the random number generator.
    vocab_exclude_list : List[str]
        List of words to force exclude from the vocabulary.
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        load_saved_model: bool = False,
        checkpoint_path: Optional[str] = None,
        cores: int = cpu_count(),
        min_word_count: int = 10,
        window: int = 5,
        vector_size: int = 300,
        sample: float = 6e-5,
        min_alpha: float = 0.007,
        negative: int = 20,
        epochs: int = 30,
        seed: int = 42,
        vocab_exclude_list: List[str] = [],
    ):
        """Constructor method
        """
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields
        self.load_saved_model = load_saved_model
        self.checkpoint_path = checkpoint_path
        self.cores = cores
        self.min_word_count = min_word_count
        self.window = window
        self.vector_size = vector_size
        self.sample = sample
        self.min_alpha = min_alpha
        self.negative = negative
        self.epochs = epochs
        self.seed = seed
        self.vocab_exclude_list = vocab_exclude_list
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }
        self._artifacts_write_config: EmbeddingGeneratorArtifactsWriteConfig = {
            "vocab_format": "json",
            "vocab_format_args": {},
            "embedding_matrix_format": "npy",
            "embedding_matrix_format_args": {},
        }

    def set_write_config(
        self,
        data_config: Optional[DataWriteConfig] = None,
        artifacts_config: Optional[EmbeddingGeneratorArtifactsWriteConfig] = None,
    ):
        """Overwrite the default file writing configurations"""

        if data_config:
            self._data_write_config = data_config
        if artifacts_config:
            self._artifacts_write_config = artifacts_config

    @abstractmethod
    def run(self):
        """Abstract method
        """
        pass


class CPUEmbeddingGenerator(EmbeddingGenerator):
    """The embedding generator provides an interface to create
    word embeddings or vector representations of words (tokens).

    The embedding generator uses the Word2Vec model from the Gensim library.

    Note
    ----
    This implementation of the EmbeddingGenerator is specific for CPU computation.

    Attributes
    ----------

        fields : Union[str, List[str]]
            The name of the column(s) containing text to be considered
            in the vocab and used in Word2Vec.
        load_saved_model : bool
            If True, use a pre-trained Word2Vec model.
        checkpoint_path : str
            Path to the Word2Vec model checkpoint.
        cores : int
            Number of cores to run the Word2Vec model on.
        min_word_count : int
            Ignores all words with total frequency lower than this.
        window : int
            Maximum distance between the current and predicted word.
        vector_size : int
            Output embedding size.
        sample : float
            The threshold for configuring which high-frequency
            words are randomly downsampled, use range (0, 1e-5).
        min_alpha : float
            Learning rate will linearly drop to min_alpha
            as training progresses.
        negative : int
            If > 0, negative sampling will be used.
        epochs : int
            Total number of iterations of all training data in
            the training of the Word2Vec model.
        seed : int
            Seed for the random number generator. For deterministic
            run, you need thread = 1 (aka CPU core)
            and PYTHONHASHSEED.
        vocab_exclude_list : List[str]
            Provide a list of tokens that may be present in the
            text that you would like to exclude from the vocab and
            from Word2Vec.
    """

    def __init__(self, *args, **kwargs):
        """Constructor method
        """
        super().__init__(*args, **kwargs)

        self.w2v_model = None
        self.embedding_matrix = None
        self.id_to_token = None

    def run(self, data: pa.Table, artifacts: dict) -> Tuple[pa.Table, dict]:
        """Runs a CPU-based embedding generator run method based on the
        configuration used to create the object of the CPUEmbeddingGenerator
        class

        Parameters
        ----------

        data : pyarrow.Table
            A pyarrow Table containing at least one list column containing text.
        artifacts : dict
            Artifacts are not consumed in this run method, but must be
            received in the method to operate correctly in the pipeline run method.

        Returns
        -------

        Tuple[pyarrow.Table, dict]
            The first position is a pyarrow.Table of pre-tokenized data.
            The second position is a dictionary of artifacts. The dict will
            contain keys for "embedding_matrix" and "id_to_token".
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_list_str_cols(fields=self.fields, data=data)

        # Use the Polars library to convert a text column into a list of lists
        # [each list here is a set of tokens]. The list of lists is called
        # df_words_lists. If there are multiple text columns, the df_words_list
        # is extended with contents of rows of the additional columns.
        text_data = pl.from_arrow(data.select(self.fields))
        df_words_lists = []
        for field in self.fields:
            df_words_lists.extend(text_data[field].drop_nulls().to_list())

        if self.load_saved_model and isfile(self.checkpoint_path):
            self.w2v_model = Word2Vec.load(self.checkpoint_path)
        else:
            # Configure Word2Vec
            self.w2v_model = Word2Vec(
                min_count=self.min_word_count,
                window=self.window,
                vector_size=self.vector_size,
                sample=self.sample,
                min_alpha=self.min_alpha,
                negative=self.negative,
                workers=self.cores,
                seed=self.seed,
            )

            # Word2Vec vocabulary trim rule
            def vocab_trim_rule(word: str, count: int, min_count: int):
                if count < min_count:
                    return RULE_DISCARD
                elif word in self.vocab_exclude_list:
                    return RULE_DISCARD
                else:
                    return RULE_DEFAULT

            # Train Word2Vec Model
            self.w2v_model.build_vocab(df_words_lists, trim_rule=vocab_trim_rule)
            self.w2v_model.train(
                df_words_lists,
                total_examples=self.w2v_model.corpus_count,
                epochs=self.epochs,
            )
        # If a checkpoint path was set, the word2vec model will be saved
        if self.checkpoint_path:
            self.w2v_model.save(self.checkpoint_path)

        # Sort the vocabulary
        sorted_tokens = sorted(self.w2v_model.wv.index_to_key)
        sorted_embeddings = []

        for token in sorted_tokens:
            sorted_embeddings.append(
                self.w2v_model.wv.vectors[self.w2v_model.wv.key_to_index[token]]
            )

        # Mapping between an index in the embedding matrix and a token
        self.id_to_token = {
            token_id: token
            for token_id, token in enumerate(["<pad>"] + sorted_tokens + ["<unk>"])
        }

        # Create embedding matrix with embeddings for <pad> and <unk> tokens
        pad_token_emb = np.zeros((1, self.vector_size))
        unkn_token_emb = np.random.random((1, self.vector_size)) * 0.01
        self.embedding_matrix = np.append(
            np.append(pad_token_emb, sorted_embeddings, axis=0), unkn_token_emb, axis=0
        )

        # Set up the artifacts dict to return
        produced_artifacts = {
            "embedding_matrix": self.embedding_matrix,
            "id_to_token": self.id_to_token,
        }

        return (data, produced_artifacts)

    def write_artifacts(self, write_path: str, artifacts: dict) -> None:
        """Write the artifacts produced by the embedding generator.

        Parameters
        ----------
        write_path : str
            Path is a directory where files will be written.
        artifacts : dict
            Artifacts is a dictionary of artifacts produced in this step.
            Expected keys are: "id_to_token" and "embedding_matrix".
        """
        id_to_token = artifacts["id_to_token"]
        # Add the filename with the appropriate filetype to the write path
        id_to_token_path = os.path.join(
            write_path,
            (f"id_to_token" f'.{self._artifacts_write_config["vocab_format"]}'),
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=id_to_token,
            path=id_to_token_path,
            format=self._artifacts_write_config["vocab_format"],
            **self._artifacts_write_config["vocab_format_args"],
        )

        embedding_matrix = artifacts["embedding_matrix"]
        # Add the filename with the appropriate filetype to the write path
        embedding_matrix_path = os.path.join(
            write_path,
            (
                f"embedding_matrix"
                f'.{self._artifacts_write_config["embedding_matrix_format"]}'
            ),
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=embedding_matrix,
            path=embedding_matrix_path,
            format=self._artifacts_write_config["embedding_matrix_format"],
            **self._artifacts_write_config["embedding_matrix_format_args"],
        )

    def get_parameters(self) -> dict:
        """Retrieve the embedding generator object configuration.

        Returns
        -------
        dict
            A dictionary representation of the EmbeddingGenerator object's attributes.
        """

        params = vars(self).copy()
        params["vocab_size"] = len(params["id_to_token"].keys())
        params.pop("id_to_token")
        params.pop("embedding_matrix")
        params["w2v_model"] = str(type(self.w2v_model))

        return params



"""Library of pre-defined regular expression substitution pairs."""

from bardi.nlp_engineering.regex_library.regex_set import RegexSubPair


# 0
def get_escape_code_regex() -> RegexSubPair:
    """Matches escape codes such as \\x0d, \\x0a, etc.

    Returns
    -------

    RegexSubPair
         {regex pattern, replacement string}

    Example
    -------

    Input string: ::
    
        Codes\\x0d\\x0a\\x0d \\r30

    Output string: ::
    
        Codes      30
    """

    regex_sub_pair = {"regex_str": r"(\\x[0-9A-Fa-f]{2,})|\\[stepr]", "sub_str": " "}
    return regex_sub_pair


# 1
def get_whitespace_regex() -> RegexSubPair:
    """Matches any new line, carriage return tab and multiple
    spaces and replaces it with a single space.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
        INVASIVE:\\nNegative    IN SITU:\\nN/A  IN \\tThe result\ \r 

    Output string: ::
    
        INVASIVE: Negative IN SITU: N/A IN The result
    """

    regex_sub_pair = {"regex_str": r"[\r\n\t]|\s{2,}", "sub_str": " "}
    return regex_sub_pair


# 2
def get_urls_regex() -> RegexSubPair:
    """Matches a url and replaces it
    with a URLTOKEN.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
        Source: https://www.merck.com/keytruda_pi.pdf 

    Output string: ::
    
        Source: URLTOKEN 
    """
    regex_sub_pair = {
        "regex_str": r"\b(http[s]*:\/\/)[^\s]+|\b(www\.)[^\s]+",
        "sub_str": " URLTOKEN ",
    }
    return regex_sub_pair


# 3
def get_special_punct_regex() -> RegexSubPair:
    """Matches a set of chosen punctuation symbols \\_,();[]#{}*"'~?!|^
    and replaces them with a single space.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        wt-1, ck-7 (focal) negative; [sth] ab|cd"

    Output string: ::
    
        wt-1  ck-7  focal  negative   sth  ab cd
    """
    regex_sub_pair = {
        "regex_str": r'[\\\_,\(\);\[\]#{}\*"\'\~\?!\|\^`]',
        "sub_str": " ",
    }
    return regex_sub_pair


# 4
def get_multiple_punct_regex() -> RegexSubPair:
    """Matches multiple occurences of symbols like -, .
    and _ replaces them with a single space.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
   
    Input string: ::
    
        -----this is report ___ signature

    Output string: ::
    
         this is report   signature
    """
    regex_sub_pair = {"regex_str": r"[\-\.:\/\_]{2,}", "sub_str": " "}
    return regex_sub_pair


# 5
def get_angle_brackets_regex() -> RegexSubPair:
    """Matches a content between matching angle brackets,
    keeps the content only, removes the brackets.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        <This should be fixed> But not this >90

    Output string: ::
    
        This should be fixed But not this >90
    """
    regex_sub_pair = {"regex_str": r"<(.*?)>", "sub_str": r" \1 "}
    return regex_sub_pair


# 6
def get_percent_sign_regex() -> RegexSubPair:
    """Matches the % sign and replaces
    it with a word 'percent'.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
        strong intensity >95%

    Output string: ::
    
        strong intensity >95 percent 
    """
    regex_sub_pair = {"regex_str": r"%", "sub_str": " percent "}
    return regex_sub_pair


# 7
def get_leading_digit_punctuation_regex() -> RegexSubPair:
    """Matches numeric digits at the start of a word,
    followed by punctuation and additional characters.
    Proceeds to eliminate the punctuation and inserts
    a space between the digits an the word.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)

    Example
    --------

    Input string: ::

        13-unremarkable 1-e 22-years 
        
    Output string: ::
    
        13 unremarkable   1 e   22 years  
    """
    regex_sub_pair = {
        "regex_str": r"(\b\d{1,})([\-\.:])([a-z]+)",
        "sub_str": r" \1 \3 ",
    }
    return regex_sub_pair


# 8
def get_leading_punctuation_regex() -> RegexSubPair:
    """Matches leading punctuation and removes it.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        -3a -anterior -result- :cassette

    Output string: ::
    
        3a  anterior  result-  cassette     
    """
    regex_sub_pair = {"regex_str": r"(\s[\.:\-\\])([^\s]+)", "sub_str": r" \2 "}
    return regex_sub_pair


# 9
def get_trailing_punctuation_regex() -> RegexSubPair:
    """Matches trailing punctuation and removes it.

    Returns
    -------
        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
         -3a -anterior -result- :cassette

    Output string: ::
    
         -3a -anterior  -result :cassette 
    """
    regex_sub_pair = {"regex_str": r"([^\s]+)([\.:\-\\]\s)", "sub_str": r" \1 "}
    return regex_sub_pair


# 10
def get_words_with_punct_spacing_regex() -> RegexSubPair:
    """Matches words with hyphen, colon or period and splits them.
    Requires the words to be at least two characters in length
    to avoid splitting words like ph.d.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        this-that her-2 tiff-1k description:gleason

    Output string: ::
    
        this that her-2 tiff-1k description gleason 
    """
    regex_sub_pair = {
        "regex_str": r"([a-z0-9]{2,})([\-:\.])([a-z]{2,})",
        "sub_str": r"\1 \3",
    }
    return regex_sub_pair


# 11
def get_math_spacing_regex() -> RegexSubPair:
    """Matches "math operators symbols" like ><=%:
    and adds spaces aroud them.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        This is >95% 3+3=8  6/7

    Output string: ::
    
        This is  > 95 %  3 + 3 = 8  6 / 7
    """
    regex_sub_pair = {"regex_str": r"([><=+%\/&:])", "sub_str": r" \1 "}
    return regex_sub_pair


# 12
def get_dimension_spacing_regex() -> RegexSubPair:
    """Matches digits and x and adds spaces between them.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        measuring 1.3x0.7x0.1 cm

    Output string: ::
    
        measuring 1.3 x 0.7 x 0.1 cm
    """
    regex_sub_pair = {"regex_str": r"(\d+[.\d]*)([x])", "sub_str": r"\1 \2 "}
    return regex_sub_pair


# 13
def get_measure_spacing_regex() -> RegexSubPair:
    """Matches measurements in mm, cm and ml provides
    proper spacing between the digits and measure.
    Also provides specing between 11th -> 11 th.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        10mm histologic type 2 x 3cm. this is 3.0-cm

    Output string: ::
    
        10 mm  histologic type 2 x 3 cm . this is 3.0 cm  
    """
    regex_sub_pair = {"regex_str": r"(\d+)[-]*([cpamt][mlhc])", "sub_str": r"\1 \2 "}
    return regex_sub_pair


# 14
def get_cassettes_spacing_regex() -> RegexSubPair:
    """Matches patterns like 5e-6f
    and adds spaces around them.

    Returns
    -------
        RegexSubPair - (regex pattern, replacement string)

    Examples
    --------

    Input string: ::
    
        3e-3f

    Output string: ::
    
         3e - 3f 
    """
    regex_sub_pair = {
        "regex_str": r"(\d{1,2}[a-z])(-)(\d{1,2}[a-z])|([a-z]\d{1,2})(-)([a-z]\d{1,2})",
        "sub_str": r"\1 \2 \3 ",
    }
    return regex_sub_pair


# 15
def get_dash_digits_spacing_regex() -> RegexSubPair:
    """Matches dashes around digits and adds
    spaces around the dashes.

    Returns
    -------
        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
        right 1:30-2:30 1.5-2.0 cm 0.9 cm for the 7-6

    Output string: ::
    
        right 1:30 - 2:30 1.5 - 2.0 cm 0.9 cm for the 7 - 6
    """
    regex_sub_pair = {
        "regex_str": r"( [\d+]*[\.:]*\d+\s*)(-)(\s*[\d+]*[\.:]*\d+)",
        "sub_str": r"\1 \2 \3",
    }
    return regex_sub_pair


# 16
def get_literals_floats_spacing_regex() -> RegexSubPair:
    """Matches character followed by a float and a word.
    This is a common formating problem.
    e.g. r18.0admission ->  r18.0 admission

    Returns
    -------
        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        r18.0admission diagnosis: bi n13.30admission

    Output string: ::
    
         r18.0 admission diagnosis: bi n13.30 admission 
    """
    regex_sub_pair = {
        "regex_str": r"([a-z]{1,2})(\d+\.\d+)([a-z]+)",
        "sub_str": r"\1\2 \3",
    }
    return regex_sub_pair


# 17
def get_fix_pluralization_regex() -> RegexSubPair:
    """Matches s character after a word
    and attaches it back to the word. This restores
    plural nouns demages by removed punctuation.

    Returns
    -------
        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
        specimen s code s

    Output string: ::
    
         specimens codes 
    """
    regex_sub_pair = {"regex_str": r"(\b[a-z]+)(\s+)([s]\s)", "sub_str": r"\1\3"}
    return regex_sub_pair


# 18
def get_digits_words_spacing_regex() -> RegexSubPair:
    """Matches digits that are attached to the
    beginning of a word.

    Returns
    -------
        
        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
        9837648admission

    Output string: ::
    
        9837648 admission 
    """
    regex_sub_pair = {"regex_str": r"(\s\d{1,})([a-z]{2,}\s)", "sub_str": r"\1 \2"}
    return regex_sub_pair


# 19
def get_phone_number_regex() -> RegexSubPair:
    """Matches any phone number that consists
    of 10 digits with delimeters.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        Ph: (123) 456 7890. It is (123)4567890.

    Output string: ::
    
        Ph:  PHONENUMTOKEN . It is  PHONENUMTOKEN .
    """
    regex_sub_pair = {
        "regex_str": r"\(*\d{3}\)*[-, ]*\d{3}[-, ]*\d{4}",
        "sub_str": " PHONENUMTOKEN ",
    }
    return regex_sub_pair


# 20
def get_dates_regex() -> RegexSubPair:
    """Matches dates of specified formats.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
       co: 03/09/2021 1015 completed: 03/10/21 at 3:34.

    Output string: ::
    
        co:  DATETOKEN completed:  DATETOKEN .
    """
    regex_list = [
        r"\d{1,2}\s*[\/,-\.]\s*\d{1,2}\s*[\/,-\.]\s*\d{2,4}\s*[at\s\-]*[\d{1,2}\s*[:\s*\d{1,2}]+]*(?:\s*[pa][m])*",
        r"\b(?:jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s*\d{1,2}\s*\d{2,4}",
        r"\b\d{1,2}\s*(?:jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s*\d{2,4}",
        r"\d{1,2}-(?:jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)-\d{2}\s*\d{1,2}[:\d{1,2}]+(?:\s*[pa][m])",
    ]
    consolidated_regex = f"{'|'.join(regex_list)}"
    regex_sub_pair = {"regex_str": consolidated_regex, "sub_str": " DATETOKEN "}
    return regex_sub_pair


# 21
def get_time_regex() -> RegexSubPair:
    """Matches time of format 11:20 am
    or 1.30pm or 9:52:07AM.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
        at 11:12 pm or 11.12am

    Output string: ::
    
        at  TIMETOKEN  or  TIMETOKEN  
    """
    regex_list = [
        r"(\d{1,2}\s*([:.]\s*\d{2}){1,2}\s*[ap]\.*[m]\.*)",
        r"\d{2}\s*[ap]\.*[m]\.*",
        r"[0-2][0-9]:[0-5][1-9]",
    ]
    consolidated_regex = f"{'|'.join(regex_list)}"
    regex_sub_pair = {"regex_str": consolidated_regex, "sub_str": " TIMETOKEN "}
    return regex_sub_pair


# 22
def get_address_regex() -> RegexSubPair:
    """Matches any address of format
    num (street name) in 1 to 6 words 2-letter state
    and short or long zip code.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        1034 north 500 west provo ut 84604-3337

    Output string: ::
    
          ADDRESSTOKEN  
    """
    regex_sub_pair = {
        "regex_str": r"\d+\s([0-9a-z.]+[\s,]+){1,6}[a-z]{2}[./\s+]*\d{5}(-\d{4})*",
        "sub_str": " ADDRESSTOKEN ",
    }
    return regex_sub_pair


# 23
def get_dimensions_regex() -> RegexSubPair:
    """Matches 2D or 3D dimension measurements
    and adds spaces around them.

    Returns
    -------
        RegexSubPair - (regex pattern, replacement string)

    Example
    -------

    Input string: ::
    
        3.5 x 2.5 x 9.0 cm and 33 x 6.5 cm

    Output string: ::
    
          DIMENSIONTOKEN  cm and  DIMENSIONTOKEN  cm
    """
    regex_list = [
        r"\d+\.*\d*\s*x\s*\d+\.*\d*\s*x\s*\d+\.*\d*",
        r"\d+\.*\d*\s*x\s*\d+\.*\d*",
    ]
    consolidated_regex = f"{'|'.join(regex_list)}"
    regex_sub_pair = {"regex_str": consolidated_regex, "sub_str": " DIMENSIONTOKEN "}
    return regex_sub_pair


# 24
def get_specimen_regex() -> RegexSubPair:
    """Matches marking of a pathology speciman.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        for s-21-009345 sh-22-0011300

    Output string: ::
    
         for  SPECIMENTOKEN   SPECIMENTOKEN  
    """
    regex_sub_pair = {
        "regex_str": r"[a-z]{1,3}[-]*\d{2}[-]\d{3,}[-]*",
        "sub_str": " SPECIMENTOKEN ",
    }
    return regex_sub_pair


# 25
def get_decimal_segmented_numbers_regex() -> RegexSubPair:
    """Matches combinations of digits and periods or dashes.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
         1.78.9.87

    Output string: ::
    
          DECIMALSEGMENTEDNUMBERTOKEN  
    """
    regex_sub_pair = {
        "regex_str": r"\d+[\.\-]\d+([\.\-]\d+)+",
        "sub_str": " DECIMALSEGMENTEDNUMBERTOKEN ",
    }
    return regex_sub_pair


# 26
def get_large_digits_seq_regex() -> RegexSubPair:
    """Matches large sequences of digits and replaces it.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
         456123456 
    
    Output string: ::
    
         DIGITSEQUENCETOKEN 
    """
    regex_sub_pair = {"regex_str": r"\s\d{3,}\s", "sub_str": " DIGITSEQUENCETOKEN "}
    return regex_sub_pair


# 27
def get_large_float_seq_regex() -> RegexSubPair:
    """Matches large floats and replace them.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
        456 123456.783

    Output string: ::
    
         456 LARGEFLOATTOKEN  
    """
    regex_sub_pair = {"regex_str": r"\s\d{2,}\.\d{1,}", "sub_str": " LARGEFLOATTOKEN "}
    return regex_sub_pair


# 28
def get_trunc_decimals_regex() -> RegexSubPair:
    """Matches floats and keeps only first decimal.

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::

        1.78  9.87 - 8.99

    Output string: ::
    
        1.7  9.8 - 8.9 
    """
    regex_sub_pair = {"regex_str": r"\s(\d+)(\.)(\d)(\d+)*\s", "sub_str": r" \1\2\3 "}
    return regex_sub_pair


# 29
def get_cassette_name_regex() -> RegexSubPair:
    """Matches cassettes markings of
    the specified format:

    Returns
    --------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------

    Input string: ::
    
        block:  1-e 

    Output string: ::
    
        block:  CASSETTETOKEN 
    """
    regex_list = [
        r"\s\d{1,2}[\-]*[a-z]{1,2}\s",
        r"\b[a-z][\-]*\d{1}\s",
        r"\s[a-z]\d{1,2}-\d{1,2}\s",
    ]
    consolidated_regex = f"{'|'.join(regex_list)}"
    regex_sub_pair = {"regex_str": consolidated_regex, "sub_str": " CASSETTETOKEN "}
    return regex_sub_pair


# 30
def get_duration_regex() -> RegexSubPair:
    """Matches duration specimen was treated:

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
        32d0909091

    Output string: ::
    
         DURATIONTOKEN 
    """
    regex_sub_pair = {
        "regex_str": r" \d{1,2}d\d{6,9}[.\s]*",
        "sub_str": " DURATIONTOKEN ",
    }
    return regex_sub_pair


# 31
def get_letter_num_seq_regex() -> RegexSubPair:
    """Matches a character followed directly
    by 6 to 10 digits:

    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
        c001234567

    Output string: ::
    
         LETTERDIGITSTOKEN 
    """
    regex_sub_pair = {
        "regex_str": r"\b[a-z]\d{6,10}[.\s]*",
        "sub_str": " LETTERDIGITSTOKEN ",
    }
    return regex_sub_pair


#   LAST REGEX that removed additional spaces
def get_spaces_regex() -> RegexSubPair:
    """Matches additional spaces (artifact of applying
    other regex), matches not needed periods that can be removed.

    
    Returns
    -------

        RegexSubPair - (regex pattern, replacement string)
    
    Example
    -------
    
    Input string: ::
    
        located around lower arm specimen   date

    Output string: ::
    
        located around lower arm specimen date
    """
    regex_sub_pair = {"regex_str": r"\s{2,}|\\n", "sub_str": " "}
    return regex_sub_pair



"""Dataset class definition and data handler functions for loading datasets from various sources"""

from __future__ import annotations

import json
from datetime import datetime
from typing import List, Union

import duckdb
import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.csv as csv
import pyarrow.dataset as ds
import pyarrow.feather as feather
import pyarrow.parquet as pq
from pyarrow import orc

from bardi.data.utils.pyarrow_utils import chunk_pyarrow_table


class Dataset:
    """A dataset object handles data in the form of columns and rows

    Under the hood it uses a PyArrow Table as it is a modern and efficient
    starting point for both CPU & GPU workflows.

    Attributes
    ----------

    data : PyArrow.Table | List[PyArrow.Table]
        The data table
    origin_query : str
        If a SQL data source was used by a data_handler function, the SQL query is
        recorded here for reproducibility and data provenance.
    origin_file_path : str
        If a file was used as the data source by a data_handler function, the filepath
        is recorded for reproducibility and data provenance.
    origin_format : str
        The format of the data source
    origin_row_count : int
        The total row count of the original dataset
    """

    def __init__(self):
        self.date = str(datetime.now())
        self.data = None
        self.origin_query = None
        self.origin_file_path = None
        self.origin_format = None
        self.origin_row_count = None

    def get_parameters(self) -> dict:
        params = vars(self).copy()
        params["data"] = self.data.column_names
        params["origin_query"] = (
            " ".join(self.origin_query.split()) if self.origin_query else "None"
        )
        return params


# ======= Loaders ========


def from_file(
    source: Union[str, List[str]], format: str, min_batches: int = None, *args, **kwargs
) -> Dataset:
    """Create a bardi Dataset object from a file source

    Accepted file types are: parquet, ipc, arrow, feather, csv, and orc

    The function utilizes PyArrow's dataset API to read files, and thus all
    keyword arguments available for its API are available here.

    Parameters
    ----------

    source : str, List[str]
        Path to a single file, or list of paths
    format : str
        Currently ["parquet", "ipc", "arrow", "feather", "csv", "orc"] are
        supported
    min_batches : int
        An integer number used to split the data into this amount of smaller
        tables for distribution to worker nodes in distributed computing
        environments.

    Returns
    -------

    bardi.data.data_handlers.Dataset
        bardi Dataset object with the data attribute referencing the data that
        was supplied after conversion to a PyArrow Table.

    Raises
    ------

    ValueError if the supplied file path does not contain a filetype of an
        accepted format.
    """

    format = format.lower()
    accepted_formats = ["parquet", "ipc", "arrow", "feather", "csv", "orc"]

    if format not in accepted_formats:
        raise ValueError(
            "The format given is not supported. " f"Expected one of {accepted_formats}"
        )
    else:
        # Create a PyArrow Dataset object from the file(s)
        # This step doesn't actually read the data yet, but just references it
        data = ds.dataset(source=source, format=format, *args, **kwargs)
        table = data.to_table()

        # Determine the starting number of rows in the data table
        row_count = table.num_rows

        # Create a bardi Dataset object which will reference the data and
        # additional metadata
        dataset_obj = Dataset()
        dataset_obj.origin_row_count = row_count
        dataset_obj.origin_file_path = source
        dataset_obj.origin_format = format

        if min_batches:
            # If a min_batches is specified, data from the file(s) are split
            # into a list of smaller PyArrow Tables
            tables = chunk_pyarrow_table(
                data=table, row_count=row_count, min_batches=min_batches
            )

            # Setting the bardi Dataset object's data references to the list
            # of PyArrow Tables
            dataset_obj.data = tables

        else:
            # If min_batches is not specified, all of the data is returned in
            # a single PyArrow Table
            dataset_obj.data = table

        return dataset_obj


def from_duckdb(path: str, query: str, min_batches: int = None) -> Dataset:
    """Create a bardi Dataset object using data returned from
    a custom query on a DuckDB database

    Parameters
    ----------

    path : str
        A filepath to the DuckDB database file
    query : str
        A valid SQL query adhering to DuckDB syntax specifications
    min_batches : int
        An integer number used to split the data into this amount of smaller
        tables for distribution to worker nodes.
        Number will probably align with the number of worker nodes.

    Returns
    -------

    bardi.data.data_handlers.Dataset
        bardi Dataset object with the data attribute referencing the data that
        was supplied after conversion to a PyArrow Table.
    """

    # Create a read-only connection to DuckDB database file and execute the
    # query returning a PyArrow Table
    conn = duckdb.connect(path, read_only=True)
    table = conn.execute(query).fetch_arrow_table()
    row_count = table.num_rows

    # Create a bardi Dataset object which will reference the data and
    # additional metadata
    dataset_obj = Dataset()
    dataset_obj.origin_file_path = path
    dataset_obj.origin_query = query
    dataset_obj.origin_format = "duckdb"
    dataset_obj.origin_row_count = row_count

    if min_batches:
        # If a min_batches is specified, data from the file(s) are split into a
        # list of smaller PyArrow Tables
        tables = chunk_pyarrow_table(
            data=table, row_count=row_count, min_batches=min_batches
        )

        # Setting the bardi Dataset object's data references to the list of
        # PyArrow Tables
        dataset_obj.data = tables

    else:
        # If min_batches is not specified, all of the data is returned in a
        # single PyArrow Table
        dataset_obj.data = table

    return dataset_obj


def from_pandas(df: pd.DataFrame, min_batches: int = None) -> Dataset:
    """Create a bardi dataset object from a Pandas DataFrame using the PyArrow
    function

    Parameters
    ----------

    df : Pandas DataFrame
        A Pandas DataFrame containing data intended to be passed into a bardi
        pipeline
    distributed : bool
        A flag which prompts splitting data into smaller chunks to prepare for
        later distribution to worker nodes.
        Also used to set a flag in the bardi Dataset object to direct future
        operations to be performed in a distributed manner.
    min_batches : int
        An integer number used to split the data into this amount of smaller
        tables for distribution to worker nodes.
        Number will probably align with the number of worker nodes.

    Returns
    -------

    bardi.data.data_handlers.Dataset
        bardi Dataset object with the data attribute referencing the data that
        was supplied after conversion to a PyArrow Table.
    """
    # Convert the Pandas DataFrame to a PyArrow Table
    table = pa.Table.from_pandas(df)
    row_count = table.num_rows

    # Create a bardi Dataset object which will reference the data and
    # additional metadata
    dataset_obj = Dataset()
    dataset_obj.origin_format = "pandas"
    dataset_obj.origin_row_count = row_count

    if min_batches:
        # If a min_batches is specified, data from the file(s) are split into
        # a list of smaller PyArrow Tables
        tables = chunk_pyarrow_table(
            data=table, row_count=row_count, min_batches=min_batches
        )

        # Setting the bardi Dataset object's data references to the list of
        # PyArrow Tables
        dataset_obj.data = tables

    else:
        # If min_batches is not specified, all of the data is returned in a
        # single PyArrow Table
        dataset_obj.data = table

    return dataset_obj


def from_pyarrow(table: pa.Table, min_batches: int = None) -> Dataset:
    """Create a bardi dataset object from an existing PyArrow Table

    Parameters
    ----------

    table : PyArrow Table
        A PyArrow Table containing data intended to be passed into a bardi
        pipeline
    min_batches : int
        An integer number used to split the data into this amount of smaller
        tables for distribution to worker nodes.
        Number will probably align with the number of worker nodes.

    Returns
    -------

    bardi.data.data_handlers.Dataset
        bardi Dataset object with the data attribute referencing the data that
        was supplied after conversion to a PyArrow Table.
    """

    row_count = table.num_rows

    # Create a bardi Dataset object which will reference the data and
    # additional metadata
    dataset_obj = Dataset()
    dataset_obj.origin_format = "pyarrow"
    dataset_obj.origin_row_count = row_count

    if min_batches:
        # If a min_batches is specified, data from the file(s) are split into a
        # list of smaller PyArrow Tables
        tables = chunk_pyarrow_table(
            data=table, row_count=row_count, min_batches=min_batches
        )

        # Setting the bardi Dataset object's data references to the list of
        # PyArrow Tables
        dataset_obj.data = tables

    else:
        # If min_batches is not specified, all of the data is returned in a
        # single PyArrow Table
        dataset_obj.data = table

    return dataset_obj


def from_json(json_data: Union[str, dict, List(dict)]) -> Dataset:
    """Create a bardi Dataset object from JSON data

    Parameters
    ----------

    json_data : str
        An object of name/value pairs. Names will become columns in the PyArrow
        Table.

    Returns
    -------
    bardi.data.data_handlers.Dataset
        bardi Dataset object with the data attribute referencing the data that
        was supplied after conversion to a PyArrow Table.
    """

    # Convert the JSON object into a dictionary and then load that as a single
    # row in a PyArrow Table
    if isinstance(json_data, str):
        records = [json.loads(json_data)]
    elif isinstance(json_data, dict):
        # If the JSON string was already deserialized upstream
        records = [json_data]
    elif isinstance(json_data, list):
        records = json_data
    else:
        raise TypeError(
            "json data must be provided as a serialized json str, "
            "a deserialized dictionary, or a list of dictionaries."
        )

    table = pa.Table.from_pylist(records)
    row_count = table.num_rows

    # Create a bardi Dataset object which will reference the data and
    # additional metadata
    dataset_obj = Dataset()
    dataset_obj.data = table
    dataset_obj.origin_format = "json"
    dataset_obj.origin_row_count = row_count

    return dataset_obj


# ======= Writers ========


def to_pandas(table: pa.Table) -> pd.DataFrame:
    """Return data as a pandas DataFrame
    
    Parameters
    ----------

    table : PyArrow.Table
        Table of data you want to convert to a Pandas DataFrame
    
    Returns
    -------

    pandas.DataFrame
        The same data as the input table, converted into a DataFrame
    """
    df = table.to_pandas()
    return df


def to_polars(table: pa.Table) -> pl.DataFrame:
    """Return data as a polars DataFrame
    
    Parameters
    ----------

    table : PyArrow.Table
        Table of data you want to convert to a Pandas DataFrame
    
    Returns
    -------

    polars.DataFrame
        The same data as the input table, converted into a DataFrame
    """
    df = pl.from_arrow(table)
    return df


def write_file(data: pa.Table, path: str, format: str, *args, **kwargs) -> None:
    """Write data to a file

    Note
    ----

    Only a subset of possible arguments are presented here. Additional arguments
    can be passed for specific file types.
    Reference PyArrow documentation for additional arguments.

    Parameters
    ----------

    data : PyArrow Table
    path : str
        path in filesystem where data is to be written
    format : str
        filetype in "parquet", "feather", "csv", "orc", "json", "npy"
    """

    format = format.lower()
    accepted_formats = ["parquet", "feather", "csv", "orc", "json", "npy"]

    if format not in accepted_formats:
        raise ValueError(
            "The format given is not supported. " f"Expected one of {accepted_formats}"
        )
    else:
        if format == "orc":
            orc.write_table(data, path, *args, **kwargs)
        elif format == "parquet":
            pq.write_table(data, path, *args, **kwargs)
        elif format == "feather":
            feather.write_feather(data, path, *args, **kwargs)
        elif format == "csv":
            # Convert list columns to strings
            csv_df = pl.from_arrow(data)
            schema = csv_df.schema
            csv_df = csv_df.with_columns(
                [
                    pl.format(
                        "[{}]", pl.col(field).cast(pl.List(pl.Utf8)).list.join(", ")
                    ).alias(field)
                    for field in schema.keys()
                    if schema[field] == pl.List()
                ]
            )
            data = csv_df.to_arrow()
            csv.write_csv(data, path, *args, **kwargs)
        elif format == "json":
            json_object = json.dumps(data, indent=4)
            with open(path, "w") as f:
                f.write(json_object)
        elif format == "npy":
            with open(path, "wb") as f:
                np.save(file=f, arr=data, *args, **kwargs)



from math import ceil
from typing import List

import pyarrow as pa


def chunk_pyarrow_table(
    data: pa.Table, row_count: int, min_batches: int
) -> List[pa.Table]:
    # PyArrow batch size is the maximum number of rows in a batch, so this is
    # calculated based on total number of rows and specified number of batches
    # in the batch_count keyword argument
    max_batch_size = ceil(row_count / min_batches)

    # The dataset.to_batches method returns a list of record_batches
    # (zero copy operation, just exposes data through different API)
    batches = data.to_batches(max_chunksize=max_batch_size)

    # The following generator takes each record batch and turns it into a PyArrow Table
    # creating a list of tables
    tables = [pa.Table.from_batches([batch]) for batch in batches]

    return tables



"""Split text columns into lists of tokens using simple patterns"""

from abc import abstractmethod
from typing import List, Tuple, Union, Optional

import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.utils.validations import (
    validate_pyarrow_table,
    validate_str_cols,
)
from bardi.pipeline import Step


class PreTokenizer(Step):
    """The pre-tokenizer breaks down text into smaller units
    before further tokenization is applied.

    Note
    ----

    Avoid the direct instantiation of the PreTokenizer class
    and instead instantiate one of the child classes depending
    on hardware configuration.

    Attributes
    ----------

    fields : Union[str, List[str]]
        The name of the column(s) containing text.
    split_pattern : str
        A specific pattern of characters used to divide a string
        into smaller segments or tokens.
        By default, the split is done on a single space character.
    """

    def __init__(self, fields: Union[str, List[str]], split_pattern: str = " "):
        """Constructor method
        """
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields
        self.split_pattern = split_pattern

    @abstractmethod
    def run(self):
        """Abstract method
        """
        pass


class CPUPreTokenizer(PreTokenizer):
    """The pre-tokenizer breaks down text into smaller units
    before further tokenization is applied.

    Note
    ----

    This implementation of the PreTokenizer is specific for CPU computation.

    Attributes
    ----------

    fields : Union[str, List[str]]
        The name of the column(s) containing text.
    split_pattern : str
        A specific pattern of characters used to divide a string
        into smaller segments or tokens.
        By default, the split is done on a single space character.
    """

    def __init__(self, *args, **kwargs):
        """Constructor method
        """
        super().__init__(*args, **kwargs)

    def run(
        self, data: pa.Table, artifacts: Optional[dict] = None
    ) -> Tuple[pa.Table, Union[dict, None]]:
        """Runs a CPU-based pre-tokenizer method based on the configuration
        used to create the object of the CPUPreTokenizer class.

        Parameters
        ----------

        data : pyarrow.Table
            A pyarrow Table containing at least one text column of type
            string or large_string.
        artifacts : Optional[dict]
            Artifacts are not used in this run method but must
            be received to operate correctly in the pipeline run method.

        Returns
        -------
        Tuple[pa.Table, Union[dict, None]]
            The first position is a pyarrow.Table of pre-tokenized data.
            The second position is a dictionary of artifacts. No artifacts
            are produced in this run method, so the second position will
            return None.
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_str_cols(fields=self.fields, data=data)

        # Split text fields into lists of tokens
        df = pl.from_arrow(data).with_columns(
            [
                (
                    pl.col(field)
                    .str.split(by=self.split_pattern)
                    .list.eval(pl.element().filter(pl.element() != ""))
                    .alias(field)
                )
                for field in self.fields
            ]
        )

        data = df.to_arrow()

        return (data, None)



"""Encode label columns into numerical representations"""

import os
from abc import abstractmethod
from typing import List, Tuple, TypedDict, Union, Optional

import polars as pl
import pyarrow as pa

from bardi.data import data_handlers
from bardi.nlp_engineering.utils.validations import validate_pyarrow_table
from bardi.pipeline import DataWriteConfig, Step


class LabelProcessorArtifactsWriteConfig(TypedDict):
    """Indicates the keys and data types expected in an
    artifacts write config dict for the label processor
    if overwriting the default configuration."""

    id_to_label_format: str
    id_to_label_args: Union[dict, None]


class LabelProcessor(Step):
    """The label processor encodes label columns into numerical representations and
    provides a mapping for each label to its respective representation.

    Note
    ----

    Avoid the direct instantiation of the LabelProcessor class
    and instead instantiate one of the child classes.

    Attributes
    ----------

    fields : Union[str, List[str]]
        The name of a label column of which the values
        are used to generate a standardized mapping and
        then that mapping is applied to the column.
    method : str
        Currently only a default 'unique' method is
        supported which maps each unique value in the column to an id.
    """

    def __init__(self, fields: Union[str, List[str]], method: str = "unique"):
        """Constructor method
        """
        self.fields = fields
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields
        self.method = method
        self.mapping = {}
        self.id_to_label = {}
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }
        self._artifacts_write_config: LabelProcessorArtifactsWriteConfig = {
            "id_to_label_format": "json",
            "id_to_label_format_args": {},
        }

    def set_write_config(
        self,
        data_config: Optional[DataWriteConfig] = None,
        artifacts_config: Optional[LabelProcessorArtifactsWriteConfig] = None,
    ):
        """Overwrite the default file writing configurations"""
        if data_config:
            self._data_write_config = data_config
        if artifacts_config:
            self._artifacts_write_config = artifacts_config

    @abstractmethod
    def run(self):
        """Abstract method
        """
        pass


class CPULabelProcessor(LabelProcessor):
    """The label processor creates and maps a label vocab.

    Note
    ----

    This implementation of the LabelProcessor is specific for CPU computation.

    Attributes
    ----------

    fields : Union[str, List[str]]
        The name(s) of label column(s) of which the values 
        are used to generate a standardized mapping and then that mapping is applied to the column(s).
    method : str 
        Currently only a default 'unique' method is supported which maps each unique value in the column to an id.
    mapping : dict
        Mapping dict of the form {label: id} used to convert labels in the column to ids.
    id_to_label : dict
        The reverse of mapping. Of the form {id: label} used downstream to map 
        the ids back to the original label values.
    """

    def __init__(self, *args, **kwargs):
        """Constructor method
        """
        super().__init__(*args, **kwargs)

    def run(self, data: pa.Table, artifacts: Optional[dict] = None) -> Tuple[pa.Table, dict]:
        """Run a label processor using CPU computation

        Parameters
        ----------

        data : PyArrow Table
            The data to be processed. The data must contain the
            column specified by 'field' at object creation
        artifacts : dict
            artifacts are not used in this run method, but must be received
            to operate correctly in the pipeline run method

        Returns
        -------

        Tuple[pyarrow.Table, dict]
            The first position is a pyarrow.Table of processed data.
            The second position is a dictionary of artifacts. The dict will
            contain a key for "id_to_label".

        Raises
        ------

        NotImplementedError
            A value other than 'unique' was provided for the
            label processor's method
        TypeError
            The run method was not supplied a PyArrow Table
        """
        # Perform validations
        validate_pyarrow_table(data=data)

        # Confirm that the method is passed a PyArrow Table
        produced_artifacts = {}
        for field in self.fields:
            if self.method == "unique":
                try:
                    # Get the unique labels
                    vals = (
                        pl.from_arrow(data).get_column(field).unique().sort()
                    ).to_list()

                    # Create a mapping for the labels
                    field_mapping = {label: id for id, label in enumerate(vals)}
                    # id_to_label is flipped vs what is needed for mapping
                    field_id_to_label = {
                        str(id): str(label) for label, id in field_mapping.items()
                    }
                    self.id_to_label[field] = field_id_to_label

                    df = pl.from_arrow(data).with_columns(
                        pl.col(field).map_dict(field_mapping, default=None)
                    )

                except Exception as e:
                    print(
                        f"An error occured: {e} "
                        "Processing without creating the label "
                        f"mapping for {field}"
                    )
                    artifact = {"id_to_label": {field: None}}
                    return (data, artifact)

            else:
                raise NotImplementedError("Only unique mapping is currently supported")

            data = df.to_arrow()

        # Set up the artifacts dict to return
        produced_artifacts = {"id_to_label": self.id_to_label}

        return (data, produced_artifacts)

    def write_artifacts(self, write_path: str, artifacts: dict) -> None:
        """Write the outputs produced by the label_processor

        Parameters
        ----------

        write_path : str
            Path is a directory where files will be written
        artifacts : dict
            Artifacts is a dictionary of artifacts produced in this step.
            Expected key is: "id_to_label"
        """
        id_to_token = artifacts["id_to_label"]
        # Add the filename with the appropriate filetype to the write path
        id_to_token_path = os.path.join(
            write_path,
            (f"id_to_label" f'.{self._artifacts_write_config["id_to_label_format"]}'),
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=id_to_token,
            path=id_to_token_path,
            format=self._artifacts_write_config["id_to_label_format"],
            **self._artifacts_write_config["id_to_label_format_args"],
        )

    def get_parameters(self) -> dict:
        """Retrive the label processor object configuration

        Does not return the mapping (vocab), but does return the id_to_label
        dict. This is because the mapping is just the reverse of id_to_label.

        Returns
        -------

        dict
            a dictionary representation of the splitter object's attributes
        """
        params = vars(self).copy()
        params.pop("mapping")
        params.pop("id_to_label")
        return params



from bardi.nlp_engineering.normalizer import CPUNormalizer
from bardi.nlp_engineering.pre_tokenizer import CPUPreTokenizer
from bardi.nlp_engineering.embedding_generator import CPUEmbeddingGenerator
from bardi.nlp_engineering.post_processor import CPUPostProcessor
from bardi.nlp_engineering.splitter import CPUSplitter
from bardi.nlp_engineering.splitter import NewSplit, MapSplit
from bardi.nlp_engineering.label_processor import CPULabelProcessor
from bardi.nlp_engineering.regex_library import *
from bardi.nlp_engineering.utils import validations



"""Defines a pipeline and a framework for steps to run in it"""

import os
import tracemalloc
from abc import ABCMeta, abstractmethod
from datetime import datetime
from typing import List, Literal, Tuple, TypedDict, Union

import pyarrow as pa

from bardi.data import data_handlers
from bardi.data.data_handlers import Dataset


class DataWriteConfig(TypedDict):
    data_format: str
    data_format_args: Union[dict, None]


class Step(metaclass=ABCMeta):
    """Blueprint for creating new steps for data pre-processing pipelines"""

    @abstractmethod
    def __init__(self):
        """Constructor method
        """
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config

    @abstractmethod
    def run(
        self,
        data: pa.Table,
        artifacts: dict,
    ) -> Tuple[Union[pa.Table, None], Union[dict, None]]:
        """Implement a run method in the step that will be called by the
        pipeline's run_pipeline method.

        Parameters
        ----------

        data : PyArrow.Table
            Expect to receive a PyArrow table of data
        artifacts : dict
            Expect to receive a dict of artifacts from preceding steps,
            but artifacts can be ignored in the method if values are not needed from it

        Returns
        -------
        Tuple[Union[pa.Table, None], Union[dict, None]]
            If the method performs a transformation of data then return
            the data as a PyArrow table. This will replace the pipeline
            object's processed_data attribute. If the method creates new artifacts
            besides the data, return a dictionary with keys corresponding to the name of
            the artifact and the values being the artifact's data or value
        """
        pass

    def set_write_config(self, data_config: DataWriteConfig) -> None:
        """Default implementation of the set_write_config method.

        Provide a method to customize the write configuration
        used by the step's write_outputs method if desired.

        Parameters
        ----------

        data_config : DataWriteConfig
            A typed dictionary defining the data_format (i.e., parquet, csv, etc.) 
            and any particular data_format_args available in the pyarrow API.
        """
        self._data_write_config = data_config

    def get_parameters(self) -> dict:
        """Default implementation of the get_parameters method.

        Called by the pipeline's get_parameters method. Implement a custom
        method if needed, such as removing large items from the dictionary.

        Returns
        -------

        dict
            A dictionary copy of the object's configuration.
        """
        params = vars(self).copy()
        return params

    def write_data(
        self,
        write_path: str,
        data: Union[pa.Table, None],
        data_filename: Union[str, None] = None,
    ) -> None:
        """Default implementation of the write_data method.

        Reuse existing write patterns that exist in the data handlers.

        Parameters
        ----------

        write_path : str 
            A directory where a file will be written.
        data : Union[pa.Table, None]
            PyArrow Table of data.
        data_filename : Union[str, None]
            Overwrite default file name (no filetype extension).
        """
        # Add the filename with the appropriate filetype extension to the write path
        if not data_filename:
            data_filename = f"{self.__class__.__name__}Data"
        file_path = os.path.join(
            write_path, (f'{data_filename}.{self._data_write_config["data_format"]}')
        )
        # Call the data handler write_file function
        data_handlers.write_file(
            data=data,
            path=file_path,
            format=self._data_write_config["data_format"],
            **self._data_write_config["data_format_args"],
        )

    def write_artifacts(self, write_path: str, artifacts: Union[dict, None]):
        """Default implementation of the write_artifacts method.

        Since only some steps require this method, the default behavior is 'pass'.

        Implement a custom method to write the specific artifacts produced
        in the step.

        Parameters
        ----------

        write_path : str
            A directory where a file should be written.
        artifacts : Union[dict, None]
            A dictionary of artifacts from the step that should be written to a file
        """
        pass


class Pipeline:
    """Organize Steps into a pipeline to operate on data from a Dataset"""

    def __init__(
        self,
        dataset: Dataset = None,
        write_path: str = None,
        write_outputs: Literal["pipeline-outputs", "debug", False] = "pipeline-outputs",
        data_write_config: DataWriteConfig = None,
        data_filename: str = "bardi_processed_data",
    ):
        """Create a pipeline for organizing data pre-processing steps.

        Parameters
        ----------

        dataset : bardi.data.data_handlers.Dataset
            A bardi dataset object with data to pre-process.
        write_path : str
            Directory in the file system to write outputs to.
        write_outputs : Literal["pipeline-outputs", "debug", False]
            Configuration for which outputs to write.
                *   'pipeline-outputs' will write all artifacts and final data.
                *   'debug' will write all artifacts and data from each step.
                *   False will not write any files.
        data_write_config : DataWriteConfig
            Supply a custom write configuration specifying
            file types. Default will save data as parquet files.
        data_filename : str
            Supply a filename for the final output data.
        """
        # Reference a bardi dataset object that the pipeline will operate on
        self.dataset: Dataset = dataset

        # Pipeline configuration
        self.steps: List[Step] = []
        self.num_steps = 0
        self.write_outputs = write_outputs
        self.write_path = write_path
        self.data_filename = data_filename

        # File writing configuration
        if data_write_config:
            self.data_write_config = data_write_config
        else:
            self.data_write_config = {
                "data_format": "parquet",
                "data_format_args": {"compression": "snappy", "use_dictionary": False},
            }  # Default config, but can be overwritten
        if write_outputs:
            if not os.path.isdir(self.write_path):
                raise ValueError(
                    "Pipeline write_outputs was configured to "
                    "write, however a write_path was not "
                    "correctly specified. Please provide a "
                    "path to a directory to write files into."
                )

        # Tracking information produced during pipeline execution
        self.artifacts = {}
        self.performance = {}

    def add_step(self, step: Step) -> None:
        """Adds a step object to the list of pipeline execution steps.

        Also overwrites the step's data write configuration with a consistent
        pipeline data write configuration.

        Parameters
        ----------

        step : bardi.pipeline.Step
            A bardi Step object to be added to the list of pipeline execution steps.
        """
        if isinstance(step, Step):
            step.set_write_config(data_config=self.data_write_config)
            self.steps.append(step)
            self.num_steps += 1
        else:
            raise TypeError(
                "Only objects of type Step may be added to " "a bardi pipeline."
            )

    def run_pipeline(self) -> None:
        """Calls the run and write_outputs method for each respective step
        object added to the pipeline.
        """

        if isinstance(self.dataset, Dataset):
            self.processed_data = self.dataset.data
        else:
            raise TypeError(
                "Pipeline's Dataset must reference a bardi "
                "Dataset object. Please utilize bardi's data_"
                "handlers to create a bardi Dataset object with "
                "your data."
            )

        # For each step of the pipeline, call its run method
        pipeline_start_time = datetime.now()
        for step_position, step in enumerate(self.steps, start=1):
            # Call the run method and time the execution
            step_start_time = datetime.now()
            tracemalloc.start()
            results = step.run(data=self.processed_data, artifacts=self.artifacts)
            step_max_mem = tracemalloc.get_traced_memory()[1] / 1000000
            tracemalloc.stop()
            step_end_time = datetime.now()
            step_run_time = step_end_time - step_start_time

            # Record the performance
            if self.write_outputs == "debug":
                print(f"{str(type(step))} run time: {step_run_time}")
                print(f"{str(type(step))} max memory (MB): {step_max_mem}")
            self.performance[str(type(step))] = {
                "time": str(step_run_time),
                "memory (MB)": str(step_max_mem),
            }

            if isinstance(results, tuple):
                if isinstance(results[0], pa.Table):
                    # Set the pipeline's processed_data attribute to the newest result
                    self.processed_data = results[0]

                    # Write the step's data output to a file if pipeline is configured
                    # to do so
                    if self.write_outputs == "debug":
                        step.write_data(
                            write_path=self.write_path, data=self.processed_data
                        )
                    elif (
                        step_position == self.num_steps
                        and self.write_outputs == "pipeline-outputs"
                    ):
                        step.write_data(
                            write_path=self.write_path,
                            data=self.processed_data,
                            data_filename=self.data_filename,
                        )
                if isinstance(results[1], dict):
                    # If artifacts were returned by the step, add them to
                    # the pipeline's total set of artifacts
                    self.artifacts = {**self.artifacts, **results[1]}

                    # Write the step's artifacts to files if pipeline is configured
                    # to do so
                    if self.write_outputs:
                        step.write_artifacts(
                            write_path=self.write_path, artifacts=self.artifacts
                        )
            else:
                raise TypeError(
                    "Pipeline expected step to return a tuple of "
                    "PyArrow Table of data and a dictionary of "
                    "artifacts. If the step doesn't return one of "
                    "these, that position in the tuple can be "
                    "empty, but it still needs to return a tuple."
                )

        # Record the total pipeline performance
        pipeline_end_time = datetime.now()
        pipeline_run_time = pipeline_end_time - pipeline_start_time
        if self.write_outputs == "debug":
            print(f"Pipeline run time: {pipeline_run_time}")
        self.performance[str(type(self))] = str(pipeline_run_time)

    def get_parameters(self, condensed: bool = True) -> dict:
        """Returns the parameters of the pipeline's dataset and parameters of each step.

        Parameters
        ----------

        condensed : bool
            If True, the step configuration dictionary will exclude any
            attributes set to None or False.

        Returns
        -------

        dict
            Dictionary of parameters used to configure each pipeline step object.
        """

        pipeline_params = {"dataset": {}, "steps": {}, "performance": {}}

        # Get parameters for the pipeline's dataset
        if self.dataset:
            dataset_params = self.dataset.get_parameters()
            if condensed:
                condensed_params = {}
                for attribute in dataset_params.keys():
                    if dataset_params[attribute] not in [False, None]:
                        condensed_params[attribute] = dataset_params[attribute]
                pipeline_params["dataset"][str(type(self.dataset))] = condensed_params
            else:
                pipeline_params["dataset"][str(type(self.dataset))] = dataset_params

        # Get parameters for each step in the pipeline
        for i, step in enumerate(self.steps):
            step_params = step.get_parameters()

            if condensed:
                condensed_params = {}
                for attribute in step_params.keys():
                    if step_params[attribute] not in [False, None]:
                        condensed_params[attribute] = step_params[attribute]
                pipeline_params["steps"][str(type(step))] = condensed_params
            else:
                pipeline_params["steps"][str(type(step))] = step_params

        # Get the pipeline performance
        pipeline_params["performance"] = self.performance

        return pipeline_params



"""Clean text with custom sets of regular expressions"""

from abc import abstractmethod
from typing import List, Tuple, Union, Optional

import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.regex_library.regex_lib import RegexSubPair
from bardi.nlp_engineering.utils.validations import (
    validate_pyarrow_table,
    validate_str_cols,
)
from bardi.pipeline import DataWriteConfig, Step


class Normalizer(Step):
    """Normalizer cleans and standardizes text input using regular expression
    substitutions. Lowercasing is also applied if desired.

    Note
    ----
    
    Avoid the direct instantiation of the Normalizer class and instead instantiate 
    one of the child classes depending on hardware configuration.

    Attributes
    ----------

    fields : Union[str, List[str]]
        The field or fields to be normalized.
    regex_set : List[RegexSubPair]
        List of regex substitutions to be applied.
    lowercase : Optional[bool]
        If True, lowercasing will be applied during normalization,
        defaults to True.
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        regex_set: List[RegexSubPair],
        lowercase: bool = True,
    ):
        """Constructor method
        """
        # Normalizer Configuration
        self.fields = fields
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config
        self.lowercase = lowercase
        # List of regex substitutions to be applied
        self.regex_set: List[RegexSubPair] = regex_set

    @abstractmethod
    def run(self):
        """Abstract method
        """
        pass


class CPUNormalizer(Normalizer):
    """Normalizer class for cleaning and standardizing text input using regular expression substitutions.

    Note
    ----
    This implementation of the Normalizer is specific for CPU computation.

    Attributes
    ----------
    
    fields : Union[str, List[str]]
        The name of the column(s) containing text to be normalized.
    regex_set : List[RegexSubPair]
        A list of dictionaries with keys, 'regex_str' and 'sub_str', used to
        perform regular expression substitutions of the text.
    lowercase : Optional[bool] 
        If True, lowercasing will be applied during normalization.
        Default is True.
    """

    def __init__(self, *args, **kwargs):
        """Constructor method
        """
        super().__init__(*args, **kwargs)
        # Rust uses a different syntax for regex match groups
        # than standard (uses '$' instead of backslash)
        for regex_sub_pair in self.regex_set:
            regex_sub_pair["sub_str"] = regex_sub_pair["sub_str"].replace("\\", "$")

    def run(self, data: pa.Table, artifacts: Optional[dict] = None) -> Tuple[pa.Table, dict]:
        """Run the CPU-based normalizer method based on the configuration used to create
        the object of the CPUNormalizer class.

        Parameters
        ----------

        data : pyarrow.Table
            A pyarrow Table containing at least one text column of type string
            or large_string.
        artifacts : Optional[dict]
            Artifacts are not used in this run method but must be received
            to operate correctly in the pipeline run method.

        Returns
        -------

        Tuple[pyarrow.Table, dict]
            A tuple containing the pyarrow Table of cleaned data and an empty
            dictionary.
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_str_cols(fields=self.fields, data=data)

        def implement_regex_substitutions(df: pl.DataFrame) -> pl.DataFrame:
            """Reusable function to apply each regex substitution
            normalization to each string field
            
            Parameters
            ----------

            df : pl.DataFrame
                a DataFrame containing the columns specified in the Normalizer's
                `fields` attribute
            
            Returns
            -------
            pl.DataFrame
                a DataFrame with the specified fields having its text transformed using
                the regular expression substitution pairs in the Normalizer object's 'regex_set'
                attribute
            """

            for regex_sub_pair in self.regex_set:
                df = df.with_columns(
                    [
                        pl.col(field).str.replace_all(
                            pattern=regex_sub_pair["regex_str"],
                            value=regex_sub_pair["sub_str"],
                        )
                        for field in self.fields
                    ]
                )
            return df

        # Use the Polars library to apply the normalization methods
        # to each field of the Table that is specified in self.fields
        df = (
            pl.from_arrow(data)
            .with_columns(
                [
                    pl.col(field).str.to_lowercase()
                    for field in self.fields
                    if self.lowercase
                ]
            )
            .pipe(implement_regex_substitutions)
        )

        data = df.to_arrow()

        return (data, None)



"""Curated set of regular expressions for cleaning text from pathology reports."""

from typing import List

from bardi.nlp_engineering.regex_library import regex_lib
from bardi.nlp_engineering.regex_library.regex_set import RegexSet, RegexSubPair


class PathologyReportRegexSet(RegexSet):
    """The PathologyReportRegexSet includes set of standard
    regular expression to normalize a pathology report.

    Note
    ----
    
    The set of regular expressions tailored for pathology reports
    was crafted with the understanding that dividing text based
    on punctuation often results in the loss of crucial information.
    E.g. terms like "her-2" should not be split. However, to ensure
    that the number of unique tokens remains manageable we employ a number
    of regular expression to separate some tokens around punctuation.
    E.g. :code:`22-years` becomes :code:`22 years`.
    This consideration is particularly important when 
    employing the word2vec algorithm, as an excessive number of tokens
    can impede the model's effectiveness by diluting the representation of key concepts.

    Attributes
    ----------

    convert_escape_codes: bool
        Removes escape codes such as \\x0d, \\x0a, etc.
    handle_whitespaces: bool
        Removes extra whitespaces: any new line, carriage return tab.
    remove_urls: bool
        Removes URLs found in the text that match the pattern. 
    remove_special_punct: bool 
        Removes special punctuation like (?,$).
    remove_multiple_punct: bool
        Removes duplicated punctuation.
        E.g. :code:`---`
    handle_angle_brackets: bool
        Removes angle brackets.
        E.g. :code:`<title>` becomes :code:`title`.
    replace_percent_sign: bool
        Replaces a percent sign with a 'percent' word.
    handle_leading_digit_punct: bool
        Removes punctuation when digit is attached to word.
        E.g. :code:`22-years` becomes :code:`22 years`.
    remove_leading_punct: bool
        Removes leading punctuation from words.
        E.g. :code:`-result` becomes :code:`result`.    
    remove_trailing_punct: bool
        Removes trailing punctuation from words.
        E.g. :code:`result-` becomes :code:`result`.
    handle_words_with_punct_spacing: bool
        Matches words with hyphen, colon or period and splits them.
    handle_math_spacing: bool 
        Matches "math operators symbols" like ><=%: and adds spaces aroud them.
    handle_dimension_spacing: bool
        Matches digits and x and adds spaces between them.
    handle_measure_spacing: bool
        Matches measurements in mm, cm and ml provides proper spacing between the digits and measure.
    handle_cassettes_spacing: bool
        Matches patterns like 5e-6f and adds spaces around them.
    handle_dash_digit_spacing: bool
        Matches dashes around digits and adds spaces around the dashes.
    handle_literals_floats_spacing: bool
        Matches character followed by a float and a word.
        This is a common formating problem.
        E.g. :code:`r18.0admission` becomes :code:`r18.0 admission`.
    fix_pluralization: bool
        Matches s character after a word and attaches it back to the word.
        This restores plural nouns demages by removed punctuation.
    handle_digits_words_spacing: bool
        Matches digits that are attached to the beginning of a word.
    remove_phone_numbers: bool
        Matches any phone number that consists of 10 digits with delimeters.
    remove_dates: bool
        Removes dates of prespecified format.
    remove_times: bool
        Matches time of format 11:20 am or 1.30pm or 9:52:07AM.
    remove_addresses: bool
        Matches any address of format num (street name) in 1 to 6 words 2-letter state
        and short or long zip code.
    remove_dimensions: bool
        Matches 2D or 3D dimension measurements and adds spaces around them.
    remove_specimen: bool
        Matches marking of a pathology speciman.
    remove_decimal_seg_numbers: bool
        Matches combinations of digits and periods or dashes.
        E.g. :code:` 1.78.9.87`.
    remove_large_digits_seq: bool
        Matches large sequences of digits (3 or more) and replaces it.
    remove_large_floats_seq: bool
        Matches large floats and replace them.
    trunc_decimals: bool = True
        Matches floats and keeps only first decimal.
    remove_cassette_names: bool
        Removes pathology samples' markings.
        E.g. :code:`1-e`.
    remove_duration_time: bool
        Removes duration a speciment was treated.
        E.g. :code:`32d09090301`.
    remove_letter_num_seq: bool
        Removes a character followed directly by 6 to 10 digits.
      
    """
    def __init__(
        self,
        convert_escape_codes: bool = True,  # 0
        handle_whitespaces: bool = True,  # 1
        remove_urls: bool = True,  # 2
        remove_special_punct: bool = True,  # 3
        remove_multiple_punct: bool = True,  # 4
        handle_angle_brackets: bool = True,  # 5
        replace_percent_sign: bool = True,  # 6
        handle_leading_digit_punct: bool = True,  # 7
        remove_leading_punct: bool = True,  # 8
        remove_trailing_punct: bool = True,  # 9
        handle_words_with_punct_spacing: bool = True,  # 10
        handle_math_spacing: bool = True,  # 11
        handle_dimension_spacing: bool = True,  # 12
        handle_measure_spacing: bool = True,  # 13
        handle_cassettes_spacing: bool = True,  # 14
        handle_dash_digit_spacing: bool = True,  # 15
        handle_literals_floats_spacing: bool = True,  # 16
        fix_pluralization: bool = True,  # 17
        handle_digits_words_spacing: bool = True,  # 18
        remove_phone_numbers: bool = True,  # 19
        remove_dates: bool = True,  # 20
        remove_times: bool = True,  # 21
        remove_addresses: bool = True,  # 22
        remove_dimensions: bool = True,  # 23
        remove_specimen: bool = True,  # 24
        remove_decimal_seg_numbers: bool = True,  # 25
        remove_large_digits_seq: bool = True,  # 26
        remove_large_floats_seq: bool = True,  # 27
        trunc_decimals: bool = True,  # 28
        remove_cassette_names: bool = True,  # 29
        remove_duration_time: bool = True,  # 30
        remove_letter_num_seq: bool = True,
    ):
        # === List of regex sub pairs ===
        self.regex_set: List[RegexSubPair] = []

        # === Retrieve regular expression substitution pairs from regex_lib
        #  0 removes escapes codes
        if convert_escape_codes:
            self.regex_sub_escape_codes = regex_lib.get_escape_code_regex()
            self.regex_set.append(self.regex_sub_escape_codes)

        #  1 (regex, sub_str) for any new line, carriage return tab
        # and multiple spaces --> " "
        if handle_whitespaces:
            self.regex_sub_whitespaces = regex_lib.get_whitespace_regex()
            self.regex_set.append(self.regex_sub_whitespaces)

        #  2 removes URLs that start with https http or www
        if remove_urls:
            self.regex_sub_urls = regex_lib.get_urls_regex()
            self.regex_set.append(self.regex_sub_urls)

        #  3 matches a set of special punctuation
        # ,();[]#{}* --> " "
        if remove_special_punct:
            self.regex_sub_special_punct = regex_lib.get_special_punct_regex()
            self.regex_set.append(self.regex_sub_special_punct)

        #  4 matches multiple occurences of symbols like -, .and _
        if remove_multiple_punct:
            self.regex_sub_multiple_punct = regex_lib.get_multiple_punct_regex()
            self.regex_set.append(self.regex_sub_multiple_punct)

        #  5(regex, sub_str) removes angle brackets
        # <THIS IS INSIDE> --> THIS IS INSIDE
        if handle_angle_brackets:
            self.regex_sub_angle_brackets = regex_lib.get_angle_brackets_regex()
            self.regex_set.append(self.regex_sub_angle_brackets)

        #  6 (regex, sub_str) replaces % for a percent word 56% --> 56 PERCENT
        if replace_percent_sign:
            self.regex_sub_percent_sign = regex_lib.get_percent_sign_regex()
            self.regex_set.append(self.regex_sub_percent_sign)

        #  7
        if handle_leading_digit_punct:
            self.regex_sub_leading_digit_punct = (
                regex_lib.get_leading_digit_punctuation_regex()
            )
            self.regex_set.append(self.regex_sub_leading_digit_punct)

        #  8
        if remove_leading_punct:
            self.regex_sub_leading_punct = regex_lib.get_leading_punctuation_regex()
            self.regex_set.append(self.regex_sub_leading_punct)

        #  9
        if remove_trailing_punct:
            self.regex_sub_trailing_punct = regex_lib.get_trailing_punctuation_regex()
            self.regex_set.append(self.regex_sub_trailing_punct)

        #  10
        if handle_words_with_punct_spacing:
            self.regex_sub_words_with_punct_spacing = (
                regex_lib.get_words_with_punct_spacing_regex()
            )
            self.regex_set.append(self.regex_sub_words_with_punct_spacing)

        #  11
        if handle_math_spacing:
            self.regex_sub_math_spacing = regex_lib.get_math_spacing_regex()
            self.regex_set.append(self.regex_sub_math_spacing)

        #  12
        if handle_dimension_spacing:
            self.regex_sub_dimension_spacing = regex_lib.get_dimension_spacing_regex()
            self.regex_set.append(self.regex_sub_dimension_spacing)

        #  13
        if handle_measure_spacing:
            self.regex_sub_measure_spacing = regex_lib.get_measure_spacing_regex()
            self.regex_set.append(self.regex_sub_measure_spacing)

        #  14
        if handle_cassettes_spacing:
            self.regex_sub_cassette_spacing = regex_lib.get_cassettes_spacing_regex()
            self.regex_set.append(self.regex_sub_cassette_spacing)

        #  15
        if handle_dash_digit_spacing:
            self.regex_sub_dash_spacing = regex_lib.get_dash_digits_spacing_regex()
            self.regex_set.append(self.regex_sub_dash_spacing)

        #  16
        if handle_literals_floats_spacing:
            self.regex_sub_literals_floats_spacing = (
                regex_lib.get_literals_floats_spacing_regex()
            )
            self.regex_set.append(self.regex_sub_literals_floats_spacing)

        #  17
        if fix_pluralization:
            self.regex_sub_fix_pluralization = regex_lib.get_fix_pluralization_regex()
            self.regex_set.append(self.regex_sub_fix_pluralization)

        #  18
        if handle_digits_words_spacing:
            self.regex_sub_digits_words_spacing = (
                regex_lib.get_digits_words_spacing_regex()
            )
            self.regex_set.append(self.regex_sub_digits_words_spacing)

        #  19 remove phone numbers
        if remove_phone_numbers:
            self.regex_sub_phone_numbers = regex_lib.get_phone_number_regex()
            self.regex_set.append(self.regex_sub_phone_numbers)

        #  20 remove dates and times
        if remove_dates:
            self.regex_sub_dates = regex_lib.get_dates_regex()
            self.regex_set.append(self.regex_sub_dates)

        #  21 remove time
        if remove_times:
            self.regex_sub_time = regex_lib.get_time_regex()
            self.regex_set.append(self.regex_sub_time)

        #  22 remove addresses
        if remove_addresses:
            self.regex_sub_address = regex_lib.get_address_regex()
            self.regex_set.append(self.regex_sub_address)

        #  23 dimension substitution
        if remove_dimensions:
            self.regex_sub_dimensions = regex_lib.get_dimensions_regex()
            self.regex_set.append(self.regex_sub_dimensions)

        #  24 specimen
        if remove_specimen:
            self.regex_sub_specimen = regex_lib.get_specimen_regex()
            self.regex_set.append(self.regex_sub_specimen)

        #  25
        if remove_decimal_seg_numbers:
            self.regex_sub_decimal_seg_numbers = (
                regex_lib.get_decimal_segmented_numbers_regex()
            )
            self.regex_set.append(self.regex_sub_decimal_seg_numbers)

        #  26
        if remove_large_digits_seq:
            self.regex_sub_large_digits_seq = regex_lib.get_large_digits_seq_regex()
            self.regex_set.append(self.regex_sub_large_digits_seq)

        #  27
        if remove_large_floats_seq:
            self.regex_sub_large_floats_seq = regex_lib.get_large_float_seq_regex()
            self.regex_set.append(self.regex_sub_large_floats_seq)

        #  28
        if trunc_decimals:
            self.regex_sub_trunc_decimals = regex_lib.get_trunc_decimals_regex()
            self.regex_set.append(self.regex_sub_trunc_decimals)

        #  29
        if remove_cassette_names:
            self.regex_sub_cassette_names = regex_lib.get_cassette_name_regex()
            self.regex_set.append(self.regex_sub_cassette_names)

        # 30
        if remove_duration_time:
            self.regex_sub_duration_time = regex_lib.get_duration_regex()
            self.regex_set.append(self.regex_sub_duration_time)

        # 31
        if remove_letter_num_seq:
            self.regex_sub_letter_num_seq = regex_lib.get_letter_num_seq_regex()
            self.regex_set.append(self.regex_sub_letter_num_seq)

        #  LAST condense spacing - not configurable, always executes
        self.regex_sub_spaces = regex_lib.get_spaces_regex()
        self.regex_set.append(self.regex_sub_spaces)



from distutils.core import setup

setup(
    name='bardi',
    version='0.3.1',
    author='Oak Ridge National Laboratory',
    packages=['bardi',
              'bardi.nlp_engineering',
              'bardi.data',
              'bardi.data.utils',
              'bardi.nlp_engineering.regex_library',
              'bardi.nlp_engineering.utils'],
    python_requires='>=3.8.0'
)



from typing import List, Union

from pyarrow import Table, large_list, large_string, list_, string


def validate_pyarrow_table(data: Table) -> None:
    """Confirm the data table passed is a PyArrow Table

    Keyword Arguments:
        data: PyArrow Table
            The table containing the data (passed by reference). The schema
            of this table will be checked to confirm string types. The table
            will also be confirmed to be a PyArrow Table.

    Raises:
        TypeError if the data table is not a PyArrow table
    """
    if isinstance(data, Table):
        pass
    else:
        raise TypeError("Data must be provided in a pyArrow.Table")


def validate_str_cols(fields: Union[List[str], str], data: Table) -> None:
    """Confirm the columns specified in the supplied data table are string columns

    Keyword Arguments:
        fields: Union[List[str], str]
            The field or fields that will be used in the method to
            be validated as string types

    Raises:
        TypeError if the columns supplied in the 'fields' keyword argument
        are not of PyArrow types string or large_string.
    """
    # Gather the string fields in the table
    all_pyarrow_schema_fields = [data.schema.field(name)
                                 for name in data.column_names]
    str_fields = [field.name for field in all_pyarrow_schema_fields
                  if field.type in [large_string(), string()]]

    if isinstance(fields, str):
        # If a single field was passed in to "fields" of the object,
        # check to ensure it is a string
        if fields not in str_fields:
            raise TypeError('The field indicated is not a string'
                            ' field and cannot be utilized with'
                            ' this method.')
        fields = [fields]
    elif isinstance(fields, list):
        # If a list of fields was passed in to "fields" of the object,
        # ensure they are all strings
        for field in fields:
            if field not in str_fields:
                raise TypeError(f'The field, "{field}", indicated is'
                                ' not a string field and cannot be'
                                ' utilized with this method')


def validate_list_str_cols(fields: Union[List[str], str], data: Table) -> None:
    """Confirm the columns specified in the supplied data table are list[string] columns

    Keyword Arguments:
        fields: Union[List[str], str]
            The field or fields that will be used in the method to
            be validated as string types

    Raises:
        TypeError if the columns supplied in the 'fields' keyword argument
        are not of PyArrow types list[string] or list[large_string].
    """
    # Gather the list fields in the table
    all_pyarrow_schema_fields = [data.schema.field(name)
                                 for name in data.column_names]
    list_str_fields = [field.name for field in all_pyarrow_schema_fields
                       if field.type in [large_list(large_string()),
                                         large_list(string()),
                                         list_(string()),
                                         list_(large_string())]]

    if isinstance(fields, str):
        # If a single field was passed in to "fields" of the object,
        # check to ensure it is a string
        if fields not in list_str_fields:
            raise TypeError('The field indicated is not a string'
                            ' field and cannot be utilized with'
                            ' this method.')
        fields = [fields]
    elif isinstance(fields, list):
        # If a list of fields was passed in to "fields" of the object,
        # ensure they are all strings
        for field in fields:
            if field not in list_str_fields:
                raise TypeError(f'The field, "{field}", indicated is'
                                ' not a string field and cannot be'
                                ' utilized with this method')



"""Segment the dataset into splits, such as `test, train, and val`"""

from abc import abstractmethod
from typing import Dict, List, NamedTuple, Tuple, Union

import numpy as np
import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.utils.validations import validate_pyarrow_table
from bardi.pipeline import DataWriteConfig, Step

MapSplit = NamedTuple(
    "MapSplit", [("unique_record_cols", List[str]), ("split_mapping", Dict[str, str])]
)
MapSplit.__doc__ = (
    """Specify the requirements for splitting data exactly in line with 
    an existing data split
    """
)
MapSplit.unique_record_cols.__doc__ = (
    """List of column names of which the combination forms a unique identifier in the
    dataset. This can be a single column name if that creates a unique identifier,
    but oftentimes in datasets a combination of fields are required.

    Note: This set of columns MUST create a unique record or the program will crash.
    """
)
MapSplit.split_mapping.__doc__ = (
    """`Only used for map split type.`
    A dictionary mapping where the keys are the hash of the concatenated values from
    unique_record_cols, or represented by the following pseudocode, ::
        hash(concat(*unique_record_cols))
    The values are the corresponding split value (train, test, val) or (fold1, fold2, fold3), etc.
    """
)

NewSplit = NamedTuple(
    "NewSplit",
    [
        ("split_proportions", Dict[str, float]),
        ("unique_record_cols", List[str]),
        ("group_cols", List[str]),
        ("label_cols", List[str]),
        ("random_seed", int),
    ],
)
NewSplit.__doc__ = (
    "Specify the requirements for splitting data with a new split from scratch."
)
NewSplit.split_proportions.__doc__ = (
    """`Only used for new split type.`
    Mapping of split names to split proportions. i.e., ::
        {'train': 0.75, 'test': 0.15, 'val': 0.15}
        {'fold1': 0.25, 'fold2': 0.25, 'fold3': 0.25, 'fold4': 0.25}
    Note: values must add to 1.0.
    """
)
NewSplit.unique_record_cols.__doc__ = (
    """List of column names of which the combination forms a unique identifier in the
    dataset. This can be a single column name if that creates a unique identifier,
    but oftentimes in datasets a combination of fields are required.

    Note: This set of columns MUST create a unique record or the program will crash.
    """
)
NewSplit.group_cols.__doc__ = (
    """List of column names that form a 'group' that you would like to keep in discrete
    splits. E.x., if you had multiple medical notes for a single patient,
    you may desire that all notes for a single patient end up in the same split
    to prevent potential information leakage. In this case you would provide
    something like a patient_id.
    """
)
NewSplit.label_cols.__doc__ = (
    """List of column names containing labels. Efforts are made to balance label
    distribution across splits, but this is not guaranteed.
    """
)
NewSplit.random_seed.__doc__ = (
    "Required for reproducibility. If you have no preference, try on 42 for size."
)


class Splitter(Step):
    """The splitter adds a 'split' column to the data assigning
    each record to a particular split for downstream model training.

    Note
    ----

    Avoid the direct instantiation of the Splitter class
    and instead instantiate one of the child classes.
    
    Attributes
    ----------

    split_method : Union[MapSplit, NewSplit]
        A named tuple of either MapSplit type or NewSplit
        type. Each contains a different set of values
        used to create the splitter depending upon split type
    split_type : str
        The type of split to be performed:
            * new - create a new random data split
            * map - reproduce an existing data split by mapping unique IDs
    unique_record_cols : List[str]
        List of column names of which the combination forms a unique identifier in the
        dataset. This can be a single column name if that creates a unique identifier,
        but oftentimes in datasets a combination of fields are required.

        Note: This set of columns MUST create a unique record or the program will crash.
    split_mapping : dict[str, str]
        `Only used for map split type.`
        A dictionary mapping where the keys are the hash of the concatenated values from
        unique_record_cols, or represented by the following pseudocode, ::
            hash(concat(*unique_record_cols))
        The values are the corresponding split value (train, test, val) or (fold1, fold2, fold3), etc.
    split_proportions : dict[str, float]
        `Only used for new split type.`
        Mapping of split names to split proportions. i.e., ::
            {'train': 0.75, 'test': 0.15, 'val': 0.15}
            {'fold1': 0.25, 'fold2': 0.25, 'fold3': 0.25, 'fold4': 0.25}
        Note: values must add to 1.0.
    num_splits : int
        The number of splits contained in `split_proportions`.
    group_cols : List[str]
        List of column names that form a 'group' that you would like to keep in discrete
        splits. E.x., if you had multiple medical notes for a single patient,
        you may desire that all notes for a single patient end up in the same split
        to prevent potential information leakage. In this case you would provide
        something like a patient_id.
    label_cols : List[str]
        List of column names containing labels. Efforts are made to balance label
        distribution across splits, but this is not guaranteed.
    random_seed: int
        Required for reproducibility. If you have no preference, try on `42` for size.
    """

    def __init__(
        self,
        split_method: Union[MapSplit, NewSplit],
    ):
        """Constructor method
        """
        if isinstance(split_method, MapSplit):
            self.split_type = "map"
            self.unique_record_cols = split_method.unique_record_cols
            self.split_mapping = split_method.split_mapping
        elif isinstance(split_method, NewSplit):
            self.split_type = "new"
            self.split_proportions = split_method.split_proportions
            self.num_splits = len(self.split_proportions)
            self.unique_record_cols = split_method.unique_record_cols
            self.group_cols = split_method.group_cols
            self.label_cols = split_method.label_cols
            self.random_seed = split_method.random_seed
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config

    @abstractmethod
    def run(self):
        """Abstract method
        """
        pass


class CPUSplitter(Splitter):
    """The splitter adds a 'split' column to the data assigning
    each record to a particular split for downstream model training.

    Two split types are available - creating a new random split
    from scratch and assigning previously created splits. This
    second option is helpful when running comparisons with
    other methods of data processing ensuring that splits are
    exactly the same.

    Note
    ----
    This implementation of the LabelProcessor is specific for CPU computation.
    
    To create a splitter, pass the appropriate set of parameters through a defined
    NamedTuple for the type of split you want to create. i.e., ::

        CPUSplitter(split_method=NewSplit(
            split_proportions={
                'train': 0.75,
                'test': 0.15,
                'val': 0.15
            },
            unique_record_cols=[
                'document_id'
            ],
            group_cols=[
                'patient_id_number',
                'registry'
            ],
            labels_cols=[
                'reportability'
            ],
            random_seed=42
        ))
    
    Splitter Method Named Tuples:
    
        *   :mod:`bardi.nlp_engineering.splitter.NewSplit`
        *   :mod:`bardi.nlp_engineering.splitter.MapSplit`

    Attributes
    ----------
    split_method : Union[MapSplit, NewSplit]
        A named tuple of either MapSplit type or NewSplit
        type. Each contains a different set of values
        used to create the splitter depending upon split type
    split_type : str
        The type of split to be performed:
            * new - create a new random data split
            * map - reproduce an existing data split by mapping unique IDs
    unique_record_cols : List[str]
        List of column names of which the combination forms a unique identifier in the
        dataset. This can be a single column name if that creates a unique identifier,
        but oftentimes in datasets a combination of fields are required.

        Note: This set of columns MUST create a unique record or the program will crash.
    split_mapping : dict[str, str]
        `Only used for map split type.`
        A dictionary mapping where the keys are the hash of the concatenated values from
        unique_record_cols, or represented by the following pseudocode, ::
            hash(concat(*unique_record_cols))
        The values are the corresponding split value (train, test, val) or (fold1, fold2, fold3), etc.
    split_proportions : dict[str, float]
        `Only used for new split type.`
        Mapping of split names to split proportions. i.e., ::
            {'train': 0.75, 'test': 0.15, 'val': 0.15}
            {'fold1': 0.25, 'fold2': 0.25, 'fold3': 0.25, 'fold4': 0.25}
        Note: values must add to 1.0.
    num_splits : int
        The number of splits contained in `split_proportions`.
    group_cols : List[str]
        List of column names that form a 'group' that you would like to keep in discrete
        splits. E.x., if you had multiple medical notes for a single patient,
        you may desire that all notes for a single patient end up in the same split
        to prevent potential information leakage. In this case you would provide
        something like a patient_id.
    label_cols : List[str]
        List of column names containing labels. Efforts are made to balance label
        distribution across splits, but this is not guaranteed.
    random_seed: int
        Required for reproducibility. If you have no preference, try on `42` for size.
    """

    def __init__(self, *args, **kwargs):
        """Constructor method
        """
        super().__init__(*args, **kwargs)

    def run(self, data: pa.Table, artifacts: dict = None) -> Tuple[pa.Table, dict]:
        """Runs a splitter using CPU computation based on the configuration
        used to create the object of the CPUSplitter class

        Parameters
        ----------

        data : PyArrow Table
            The data to be split
        artifacts : dict
            artifacts are not consumed in this run method, but must be
            received to operate correctly in the pipeline run method

        Returns
        -------

        Tuple(PyArrow.Table, dict)
            A tuple with:
                *   the first element holding a PyArrow Table of data
                    including the new split column
                *   the second element of the tuple intended for
                    artifacts is None

        Raises
        ------

        TypeError
            The run method was not supplied a PyArrow Table
        """
        # Perform validations
        validate_pyarrow_table(data=data)

        df = pl.from_arrow(data)

        # Mapping from an existing data split - Good for comparisons
        if self.split_type == "map":
            # Apply an existing mapping of id -> split

            # Combine and hash the columns that form a unique/distinct
            # record and map a dictionary set up as:
            # {id: split} where id is also a combined and hashed
            # combination of the columns forming a unique record
            df = (
                df.with_columns(
                    [
                        pl.concat_str([*self.unique_record_cols])
                        .hash()
                        .alias("composite_record_id")
                    ]
                )
                .with_columns(
                    [
                        pl.col("composite_record_id")
                        .cast(pl.Utf8())
                        .map_dict(self.split_mapping)
                        .alias("split")
                    ]
                )
                .drop("composite_record_id")
            )

        elif self.split_type == "new":
            # Create a new split of the data if an existing desired
            # split does not yet exist

            # Get the set of distinct groups from the overall data
            distinct_groups = df.select([*self.group_cols]).unique(maintain_order=True)
            group_count = distinct_groups.height
            # Generate a permutation of indices of distinct groups.
            np.random.seed(self.random_seed)
            permuted_indices = np.random.permutation(group_count)

            # The splits assignment has to be done with a number
            # so a mapping must be created from a number to the
            # desired split name
            split_name_maps = {
                i: key for i, key in enumerate(self.split_proportions.keys())
            }

            # Create a numpy array to hold the split labels (digits)
            split_labels = np.empty(group_count, dtype=int)
            start, end = 0, 0
            # Loop over the split proportion and assign digit
            # labels to the corresponding indices. If 15% of
            # data should be in a given split then we take a next
            # slice of permutated indices of that size and assign to them
            # a given digit label.
            for i, split_proportion in enumerate(self.split_proportions.items()):
                end = start + int(split_proportion[1] * group_count)
                split_labels[permuted_indices[start:end]] = i
                start = end
            # In case some of incides are left out due to divisibility.
            split_labels[permuted_indices[end:]] = i
            split_labels = pl.Series(split_labels)

            # Add split labels as a column to distince groups
            # mapped to assigned value back to the name of
            # the split
            splits = distinct_groups.with_columns(
                split_labels.map_dict(split_name_maps).alias("split")
            )

            # Join the split set of distinct groups back to the rest
            # of the data. If a group had more than one record in the
            # dataset, this join will ensure those records are mapped
            # to the same split as the others in the group
            df = df.join(splits, on=[*self.group_cols], how="left")

        data = df.to_arrow()

        return (data, None)



"""Define RegexSet and RegexSubPair blueprints"""

from typing import List, TypedDict


class RegexSubPair(TypedDict):
    """Dictionary used for regular expression string substitutions

    Example of a regex sub pair dictionary: ::
    
        {
            "regex_str": r"\\s",
            "sub_str": "WHITESPACE"
        }

    Attributes
    ----------

    regex_str : str 
        regular expression pattern
    sub_str : str
        replacement value for matched string
    """

    regex_str: str
    sub_str: str


class RegexSet:
    """Blueprint for creating a configurable, domain specific regular expression set
    
    Attributes
    ----------

    regex_set : List[RegexSubPair]
        a list of regular expression substitution pairs
    """

    def __init__(self):
        self.regex_set: List[RegexSubPair] = []

    def get_regex_set(self) -> List[RegexSubPair]:
        """Return the ordered set of regular expressions
        
        Returns
        -------

        List[RegexSubPair]
            a list of regular expression substitution pairs
        """
        return self.regex_set
